const SSD1306 = require('./../../constants/SSD1306');
const OLED = require('./core');

function drawText(x, y, font, text) {
    let yPosition = y;
    let xPosition = x;
    let fontHeight = font.height;
    let space = font.space;
    let index = 0;

    for (let i = 0; i < text.length; i++) {
        xPosition = (index * space + x) % SSD1306.WIDTH;
        index++;
        if ((xPosition + space) >= SSD1306.WIDTH) {
            yPosition += fontHeight + 1;
            xPosition = 0;
            x = 0;
            index = 1;
        }
        drawChar(xPosition, yPosition, font, text[i]);
    }
}

function drawChar(x, y, font, char) {
    let charBuffer = findCharBuffer(font, char);
    let byteArray = readCharBytes(charBuffer);
    let pagePosition = 0;
    let xPosition = x;
    let c = 0;

    for (let i = 0; i < byteArray.length; i++) {
        pagePosition = (Math.floor(i / font.width)) * 8;

        if (font.height <= 8 || (font.height > 8 && font.height <= 16 && !(i % 2)) || (font.height > 16 && !(i % 3))) {
            c++;
        }

        xPosition = c + x;
        for (let j = 0; j < 8; j += 1) {
            let color = byteArray[i][j];
            let yPosition = 0;
            if ((font.height > 16 && !(i % 3)) || font.height <= 8 || (font.height > 8 && font.height <= 16 && !(i % 2))) {
                yPosition += y + j + pagePosition;
                OLED.setPixel(xPosition, yPosition, color);
            } else if ((font.height > 16 && !((i - 1) % 3)) || (font.height > 8 && font.height <= 16 && (i % 2))) {
                yPosition += y + j + pagePosition + 8;
                OLED.setPixel(xPosition, yPosition, color);
            } else {
                yPosition += y + j + pagePosition + 16;
                OLED.setPixel(xPosition, yPosition, color);
            }

        }
    }
}

function findCharBuffer(font, c) {
    let cBufPosition = (c.charCodeAt(0) - font.offset) * font.width;
    return font.fontData.slice(cBufPosition, cBufPosition + font.width);
}

function readCharBytes(byteArray) {
    let bitArray = [],
        bitCharArray = [];
    for (let i = 0; i < byteArray.length; i++) {
        let byte = byteArray[i];
        for (let j = 0; j < 8; j++) {
            let bit = byte >> j & 1;
            bitArray.push(bit);
        }
        bitCharArray.push(bitArray);
        bitArray = [];
    }
    return bitCharArray;
}

function drawFillRectangle(x0, y0, w, h, color) {
    for (let i = x0; i < x0 + w; i++) {
        writeFastVLine(i, y0, h, color);
    }
}

function swap(a, b) {
    return [b, a];
}

function writeFastVLine(x, y, h, color) {
    drawLine(x, y, x, y + h - 1, color);
}

function drawLine(x0, y0, x1, y1, color) {
    let steep = Math.abs(y1 - y0) > Math.abs(x1 - x0);
    if (steep) {
        [x0, y0] = swap(x0, y0);
        [x1, y1] = swap(x1, y1);
    }

    if (x0 > x1) {
        [x0, x1] = swap(x0, x1);
        [y0, y1] = swap(y0, y1);
    }

    let dx, dy;
    dx = x1 - x0;
    dy = Math.abs(y1 - y0);

    let err = dx / 2;
    let yStep;

    if (y0 < y1) {
        yStep = 1;
    } else {
        yStep = -1;
    }

    for (; x0 <= x1; x0++) {
        if (steep) {
            OLED.setPixel(y0, x0, color);
        } else {
            OLED.setPixel(x0, y0, color);
        }
        err -= dy;
        if (err < 0) {
            y0 += yStep;
            err += dx;
        }
    }
}

function drawCircle(x0, y0, radius, color) {
    let x = radius - 1;
    let y = 0;
    let dx = 1;
    let dy = 1;
    let diameter = radius * 2;
    let decisionOver2 = dx - diameter;

    while (x >= y) {
        OLED.setPixel(x0 + x, y0 + y, color);
        OLED.setPixel(x0 + y, y0 + x, color);
        OLED.setPixel(x0 - y, y0 + x, color);
        OLED.setPixel(x0 - x, y0 + y, color);
        OLED.setPixel(x0 - x, y0 - y, color);
        OLED.setPixel(x0 - y, y0 - x, color);
        OLED.setPixel(x0 + y, y0 - x, color);
        OLED.setPixel(x0 + x, y0 - y, color);

        if (decisionOver2 <= 0) {
            y++;
            decisionOver2 += dy;
            dy += 2;
        }
        if (decisionOver2 > 0) {
            x--;
            dx += 2;
            decisionOver2 += (-diameter) + dx;
        }
    }

}

async function fontTester(font) {
    let buffer = "";
    let speed = 200;
    let size = font.fontData.length / font.width + font.offset;

    for (let i = font.offset; i < size; i++) {
        buffer += String.fromCharCode(i);
    }

    for (let i = buffer.length; i >= 0; i--) {
        OLED.clear();
        drawText(0, 0, font, buffer.slice(i), 1);
        OLED.update();
        await  sleep(speed);
    }
}

function sleep(ms) {
    return new Promise(resolve => setTimeout(resolve, ms));
}

module.exports = {
    init: OLED.init,
    update: OLED.update,
    clear: OLED.clear,
    drawText: drawText,
    fontTester: fontTester,
    drawCircle: drawCircle,
    drawLine: drawLine,
    writeFastVLine: writeFastVLine,
    drawFillRectangle: drawFillRectangle,
    setPixel: OLED.setPixel
};

