const USB = require('usb');
const common = require('./../../constants/common');
const I2C = require('./../../constants/I2C');
const SSD1306 = require('./../../constants/SSD1306');
const PID = common.PID;
const VID = common.VID;
const device = USB.findByIds(VID, PID);
const buffer = new Uint8Array((SSD1306.WIDTH * SSD1306.HEIGHT) / 8);
const DATA = {
    bmRequestType: 0x40,
    bRequest: 0,
    wValue: 0,
    wIndex: 0,
    data_or_length: 0
};

function init(vcc, refresh, bitrate) {
    if (typeof vcc === "undefined") vcc = SSD1306.SWITCHCAPVCC;
    if (typeof refresh === "undefined") refresh = SSD1306.REFRESH_MID;
    if (typeof bitrate === "undefined") bitrate = 500;

    device.open();
    setBitrate(bitrate);

    sendCmd(SSD1306.DISPLAYOFF);
    sendCmd(SSD1306.SETDISPLAYCLOCKDIV);
    sendCmd(refresh);

    sendCmd(SSD1306.SETDISPLAYOFFSET);
    sendCmd(0x0);
    sendCmd(SSD1306.SETSTARTLINE | 0x0);
    sendCmd(SSD1306.CHARGEPUMP);

    (vcc === SSD1306.EXTERNALVCC) ? sendCmd(0x10) : sendCmd(0x14);

    sendCmd(SSD1306.MEMORYMODE);
    sendCmd(0x00);
    sendCmd(SSD1306.SEGREMAP | 0x1);
    sendCmd(SSD1306.COMSCANDEC);

    sendCmd(SSD1306.SETCONTRAST);

    (vcc === SSD1306.EXTERNALVCC) ? sendCmd(0x9F) : sendCmd(0xCF);

    sendCmd(SSD1306.SETPRECHARGE);

    sendCmd(SSD1306.SETMULTIPLEX);
    sendCmd(0x3F);

    sendCmd(SSD1306.SETCOMPINS);
    sendCmd(0x12);

    sendCmd(SSD1306.SETVCOMDETECT);
    sendCmd(0x40);

    sendCmd(SSD1306.DISPLAYALLON_RESUME);
    sendCmd(SSD1306.NORMALDISPLAY);

    sendCmd(SSD1306.DISPLAYON);

    clear();
    update();
}

function update() {
    sendCmd(SSD1306.SETLOWCOLUMN | 0x0);
    sendCmd(SSD1306.SETHIGHCOLUMN | 0x0);
    sendCmd(SSD1306.SETSTARTLINE | 0x0);

    sendBuffer(SSD1306.OLED_I2C_ADDRESS, 0x40, SSD1306.BUF_SIZE, buffer);
}

function clear() {
    buffer.fill(0x00);
}

function setPixel(x, y, color) {
    if ((x < 0) || (x >= SSD1306.WIDTH) || (y < 0) || (y >= SSD1306.HEIGHT))
        return;

    let byte = 0;
    let page = Math.floor(y / 8);
    let pageShift = 0x01 << (y - 8 * page);

    (page === 0) ? byte = x : byte = x + (SSD1306.WIDTH * page);
    (color === 0) ? buffer[byte] &= ~pageShift : buffer[byte] |= pageShift;
}

function sendBuffer(address, control, BUF_SIZE, buf) {
    DATA.bmRequestType = 0x40;
    DATA.bRequest = I2C.FUNC_I2C_TX_MORE;
    DATA.wValue = address;
    DATA.wIndex = control;
    DATA.data_or_length = Buffer(buf);

    device.controlTransfer(DATA.bmRequestType, DATA.bRequest, DATA.wValue, DATA.wIndex, DATA.data_or_length, function (error) {
        if (error !== undefined) console.log(`[sendBuffer]: ${error}`);
    });
}

function sendCmd(cmd) {
    DATA.bmRequestType = 0x40;
    DATA.bRequest = I2C.FUNC_I2C_TX_SHORT;
    DATA.wValue = SSD1306.OLED_I2C_ADDRESS;
    DATA.wIndex = (((cmd & 0xff) << 8) | (0x00 & 0xff));
    DATA.data_or_length = Buffer(1);

    device.controlTransfer(DATA.bmRequestType, DATA.bRequest, DATA.wValue, DATA.wIndex, DATA.data_or_length, function (error) {
        if (error !== undefined) console.log(`[sendCmd]: ${error}`);
    });
}

function setBitrate(bitrate) {
    DATA.bmRequestType = 0xC0;
    DATA.bRequest = I2C.FUNC_I2C_BITRATE;
    DATA.wValue = bitrate;
    DATA.wIndex = 0;
    DATA.data_or_length = 254;

    device.controlTransfer(DATA.bmRequestType, DATA.bRequest, DATA.wValue, DATA.wIndex, DATA.data_or_length, function (error, data) {
        if (error !== undefined) console.log(`[setBitrate]: ${error}`);
        if (data !== undefined) console.log(`[setBitrate = ${bitrate}kHz]: ${data}`);
    });
}

module.exports = {
    init: init,
    update: update,
    clear: clear,
    setPixel: setPixel
};